
#ifndef MIGRAPHX_EXPORT_H
#define MIGRAPHX_EXPORT_H

#ifdef MIGRAPHX_STATIC_DEFINE
#  define MIGRAPHX_EXPORT
#  define MIGRAPHX_NO_EXPORT
#else
#  ifndef MIGRAPHX_EXPORT
#    ifdef migraphx_EXPORTS
        /* We are building this library */
#      define MIGRAPHX_EXPORT __attribute__((visibility("default")))
#    else
        /* We are using this library */
#      define MIGRAPHX_EXPORT __attribute__((visibility("default")))
#    endif
#  endif

#  ifndef MIGRAPHX_NO_EXPORT
#    define MIGRAPHX_NO_EXPORT __attribute__((visibility("hidden")))
#  endif
#endif

#ifndef MIGRAPHX_DEPRECATED
#  define MIGRAPHX_DEPRECATED 
#endif

#ifndef MIGRAPHX_DEPRECATED_EXPORT
#  define MIGRAPHX_DEPRECATED_EXPORT MIGRAPHX_EXPORT MIGRAPHX_DEPRECATED
#endif

#ifndef MIGRAPHX_DEPRECATED_NO_EXPORT
#  define MIGRAPHX_DEPRECATED_NO_EXPORT MIGRAPHX_NO_EXPORT MIGRAPHX_DEPRECATED
#endif

/* NOLINTNEXTLINE(readability-avoid-unconditional-preprocessor-if) */
#if 0 /* DEFINE_NO_DEPRECATED */
#  ifndef MIGRAPHX_NO_DEPRECATED
#    define MIGRAPHX_NO_DEPRECATED
#  endif
#endif

#endif /* MIGRAPHX_EXPORT_H */
